@@ -11,18 +11,22 @@
 #include <stdlib.h>
 #include <sys/resource.h>
 #include <sys/types.h>
+#include <signal.h>
 #include <sys/socket.h>
 #include <sys/ioctl.h>
 #include <net/if.h>
 
 #ifdef PSUTIL_SUNOS10
     #include "arch/solaris/v10/ifaddrs.h"
+#elif PSUTIL_AIX
+    #include "arch/aix/ifaddrs.h"
 #else
     #include <ifaddrs.h>
 #endif
 
 #if defined(PSUTIL_LINUX)
     #include <netdb.h>
+    #include <linux/types.h>
     #include <linux/if_packet.h>
 #elif defined(PSUTIL_BSD) || defined(PSUTIL_OSX)
     #include <netdb.h>
@@ -34,8 +38,98 @@
 #elif defined(PSUTIL_SUNOS)
     #include <netdb.h>
     #include <sys/sockio.h>
+#elif defined(PSUTIL_AIX)
+    #include <netdb.h>
 #endif
 
+#include "_psutil_common.h"
+
+/*
+ * Check if PID exists. Return values:
+ * 1: exists
+ * 0: does not exist
+ * -1: error (Python exception is set)
+ */
+int
+psutil_pid_exists(long pid) {
+    int ret;
+
+    // No negative PID exists, plus -1 is an alias for sending signal
+    // too all processes except system ones. Not what we want.
+    if (pid < 0)
+        return 0;
+
+    // As per "man 2 kill" PID 0 is an alias for sending the signal to
+    // every process in the process group of the calling process.
+    // Not what we want. Some platforms have PID 0, some do not.
+    // We decide that at runtime.
+    if (pid == 0) {
+#if defined(PSUTIL_LINUX) || defined(PSUTIL_FREEBSD)
+        return 0;
+#else
+        return 1;
+#endif
+    }
+
+#if defined(PSUTIL_OSX)
+    ret = kill((pid_t)pid , 0);
+#else
+    ret = kill(pid , 0);
+#endif
+
+    if (ret == 0)
+        return 1;
+    else {
+        if (errno == ESRCH) {
+            // ESRCH == No such process
+            return 0;
+        }
+        else if (errno == EPERM) {
+            // EPERM clearly indicates there's a process to deny
+            // access to.
+            return 1;
+        }
+        else {
+            // According to "man 2 kill" possible error values are
+            // (EINVAL, EPERM, ESRCH) therefore we should never get
+            // here. If we do let's be explicit in considering this
+            // an error.
+            PyErr_SetFromErrno(PyExc_OSError);
+            return -1;
+        }
+    }
+}
+
+
+/*
+ * Utility used for those syscalls which do not return a meaningful
+ * error that we can translate into an exception which makes sense.
+ * As such, we'll have to guess.
+ * On UNIX, if errno is set, we return that one (OSError).
+ * Else, if PID does not exist we assume the syscall failed because
+ * of that so we raise NoSuchProcess.
+ * If none of this is true we giveup and raise RuntimeError(msg).
+ * This will always set a Python exception and return NULL.
+ */
+int
+psutil_raise_for_pid(long pid, char *syscall_name) {
+    // Set exception to AccessDenied if pid exists else NoSuchProcess.
+    if (errno != 0) {
+        // Unlikely we get here.
+        PyErr_SetFromErrno(PyExc_OSError);
+        return 0;
+    }
+    else if (psutil_pid_exists(pid) == 0) {
+        psutil_debug("%s syscall failed and PID %i no longer exists; "
+                     "assume NoSuchProcess", syscall_name, pid);
+        NoSuchProcess("");
+    }
+    else {
+        PyErr_Format(PyExc_RuntimeError, "%s syscall failed", syscall_name);
+    }
+    return 0;
+}
+
 
 /*
  * Given a PID return process priority as a Python integer.
@@ -295,8 +389,7 @@ psutil_net_if_mtu(PyObject *self, PyObject *args) {
 error:
     if (sock != 0)
         close(sock);
-    PyErr_SetFromErrno(PyExc_OSError);
-    return NULL;
+    return PyErr_SetFromErrno(PyExc_OSError);
 }
 
 
@@ -333,13 +426,12 @@ psutil_net_if_flags(PyObject *self, PyObject *args) {
 error:
     if (sock != 0)
         close(sock);
-    PyErr_SetFromErrno(PyExc_OSError);
-    return NULL;
+    return PyErr_SetFromErrno(PyExc_OSError);
 }
 
 
 /*
- * net_if_stats() OSX/BSD implementation.
+ * net_if_stats() macOS/BSD implementation.
  */
 #if defined(PSUTIL_BSD) || defined(PSUTIL_OSX)
 
@@ -526,10 +618,9 @@ psutil_net_if_duplex_speed(PyObject *self, PyObject *args) {
 error:
     if (sock != 0)
         close(sock);
-    PyErr_SetFromErrno(PyExc_OSError);
-    return NULL;
+    return PyErr_SetFromErrno(PyExc_OSError);
 }
-#endif  // net_if_stats() OSX/BSD implementation
+#endif  // net_if_stats() macOS/BSD implementation
 
 
 /*
@@ -607,7 +698,7 @@ void init_psutil_posix(void)
     PyObject *module = Py_InitModule("_psutil_posix", PsutilMethods);
 #endif
 
-#if defined(PSUTIL_BSD) || defined(PSUTIL_OSX) || defined(PSUTIL_SUNOS)
+#if defined(PSUTIL_BSD) || defined(PSUTIL_OSX) || defined(PSUTIL_SUNOS) || defined(PSUTIL_AIX)
     PyModule_AddIntConstant(module, "AF_LINK", AF_LINK);
 #endif
 
