@@ -227,6 +227,13 @@ def wheel():
     sh("%s setup.py bdist_wheel" % PYTHON)
 
 
+@cmd
+def upload_wheels():
+    """Upload wheel files on PYPI."""
+    build()
+    sh("%s -m twine upload dist/*.whl" % PYTHON)
+
+
 @cmd
 def install_pip():
     """Install pip"""
